import os
from typing import List
from urllib.parse import urlparse

from ..request import Request
from ..response import Response
from ..types import Processor


def new(base_dir: str = os.path.expanduser('~/Downloads'), *, sub_dir_attr: str | list[str] = None,
        filename_attr: str = None) -> Processor:
    def process(res: Response) -> List[Request]:
        if sub_dir_attr is not None:
            dir_attrs = sub_dir_attr if isinstance(sub_dir_attr, list) else [sub_dir_attr]
        else:
            dir_attrs = []
        d = os.path.join(base_dir, *[str(res.request.get_attr(attr)) for attr in dir_attrs]) if dir_attrs else base_dir
        n = res.request.get_attr(filename_attr) if filename_attr else None
        if not n:
            disposition = res.headers.get('Content-Disposition', None)
            if disposition:
                if 'filename=' in disposition:
                    filename = disposition.split('filename=')[1]
                    if len(filename) > 2:
                        n = filename[1:-1]
        n = n if n else os.path.basename(urlparse(res.request.url).path)
        if not os.path.exists(d):
            os.makedirs(d)
        file = os.path.join(d, n)
        if os.path.exists(file):
            os.remove(file)
        with open(file, 'wb') as f:
            f.write(res.content)
        return []

    return process



from datetime import datetime
from typing import List

from ..types import Converter

def int_converter(default_value: int = None) -> Converter:
    return lambda x: int(x) if x else default_value


def float_converter(default_value: int = None) -> Converter:
    return lambda x: float(x) if x else default_value


def bool_converter(true_values: List[str] = None, default_value: int = None) -> Converter:
    if true_values is None:
        true_values = ['true', '1', 'y', 'yes', 'on', 't']
    return lambda x: x.lower() in true_values if x else default_value


def datetime_converter(pattern='%Y-%m-%d %H:%M:%S', default_value: datetime = None) -> Converter:
    return lambda x: datetime.strptime(x, pattern) if x else default_value



from .selector import Selector
from .. import Response
from ..types import Converter


class Header(Selector):
    def __init__(self,
                 expr: str,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Header, self).__init__(expr,
                                     multi=multi,
                                     trim=trim,
                                     filter_empty=filter_empty,
                                     format_str=format_str,
                                     converter=converter)

    def do_select(self, content: str, resp: Response) -> list[str]:
        attr = resp.request.headers.get(self.expr, None)
        if attr is None:
            return []
        return [str(c) for c in attr] if isinstance(attr, list) else [str(attr)]


def new(name: str,
        *,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Header:
    return Header(name,
                  multi=multi,
                  trim=trim,
                  filter_empty=filter_empty,
                  format_str=format_str,
                  converter=converter)



from lxml import html

from .selector import Selector
from .. import Response
from ..types import Converter


class Xpath(Selector):
    def __init__(self, expr: str,
                 selector: Selector = None,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Xpath, self).__init__(expr,
                                    selector=selector,
                                    multi=multi,
                                    trim=trim,
                                    filter_empty=filter_empty,
                                    format_str=format_str,
                                    converter=converter)

    def do_select(self, content: str, resp: Response) -> list[str]:
        return [html.tostring(x, encoding=resp.encoding).decode() if isinstance(x, html.HtmlElement) else str(x) for x
                in
                html.fromstring(content).xpath(self.expr)]


def new(expr: str,
        selector: Selector = None,
        *,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Xpath:
    return Xpath(expr, selector, multi=multi, trim=trim, filter_empty=filter_empty,
                 format_str=format_str,
                 converter=converter)



from .excel_collector import new as excel_collector
from .excel_collector import new_cell_style as excel_style
from .excel_collector import new_column as excel_column
from .logging_collector import new as logging_collector

__all__ = [
    'excel_collector',
    'excel_column',
    'excel_style',
    'logging_collector'
]



from .attr import new as attr
from .css import new as css
from .header import new as header
from .id import new as id
from .json import new as json
from .query import new as query
from .regex import new as regex
from .selector import embedded, select, link, hyperlink
from .url import new as url
from .xpath import new as xpath

__all__ = [
    'embedded',
    'select',
    'css',
    'xpath',
    'regex',
    'json',
    'attr',
    'url',
    'link',
    'hyperlink',
    'query',
    'header',
    'id'
]



import logging

from ..types import Collector, R

_logger = logging.getLogger('pyoctopus.collector.logging')


def new() -> Collector:
    def _collect(r: R) -> None:
        _logger.info(str(r))

    return _collect



from .selector import Selector
from .. import Response
from ..types import Converter


class Id(Selector):
    def __init__(self,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Id, self).__init__('',
                                 multi=multi,
                                 trim=trim,
                                 filter_empty=filter_empty,
                                 format_str=format_str,
                                 converter=converter)

    def do_select(self, content: str, resp: Response) -> list[str]:
        return [resp.request.id]


def new(*,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Id:
    return Id(
        multi=multi,
        trim=trim,
        filter_empty=filter_empty,
        format_str=format_str,
        converter=converter)



from .matcher import and_matcher, or_matcher, not_matcher, host_matcher, url_matcher, header_matcher, \
    content_type_matcher, ALL, JSON, HTML, IMAGE, PDF, WORD, EXCEL, OCTET_STREAM, MEDIA, AUDIO, VIDEO

__all__ = [
    'and_matcher', 'or_matcher', 'not_matcher', 'host_matcher', 'url_matcher', 'header_matcher',
    'content_type_matcher', 'ALL', 'JSON', 'HTML', 'IMAGE', 'PDF', 'WORD', 'EXCEL', 'OCTET_STREAM', 'MEDIA', 'AUDIO',
    'VIDEO'
]



from .selector import Selector
from .. import Response
from ..types import Converter
from bs4 import BeautifulSoup


class Css(Selector):
    def __init__(self,
                 expr: str,
                 attr=None,
                 text=False,
                 selector:
                 Selector = None,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Css, self).__init__(expr,
                                  selector=selector,
                                  multi=multi,
                                  trim=trim,
                                  filter_empty=filter_empty,
                                  format_str=format_str,
                                  converter=converter)
        self.attr = attr
        self.text = text

    def do_select(self, content: str, resp: Response) -> list[str]:
        html = BeautifulSoup(content, 'lxml')
        return [(x.attrs[self.attr] if self.attr else (x.text if self.text else x.decode())) for x in
                html.select(self.expr)]


def new(expr: str,
        attr=None,
        text=False,
        selector: Selector = None,
        *,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Css:
    return Css(expr,
               attr,
               text,
               selector,
               multi=multi,
               trim=trim,
               filter_empty=filter_empty,
               format_str=format_str,
               converter=converter)



import logging
from abc import abstractmethod
from typing import List, Any

from .. import Request, Response
from ..types import Converter, Terminable
from ..types import R

PROP_LINKS = '__result_links__'
_logger = logging.getLogger('pyoctopus.selector')


class Link:
    def __init__(self,
                 selector: 'Selector',
                 method: str = 'GET',
                 *,
                 queries: dict[str, list[str]] = None,
                 data: str = None,
                 headers: dict[str, str] = None,
                 priority: int = 0,
                 repeatable: bool = True,
                 attr_props: list[str] = None,
                 inherit: bool = False,
                 terminable: Terminable = None):
        self.selector = selector
        self.method = method
        self.queries = {} if queries is None else queries
        self.data = data
        self.headers = {} if headers is None else headers
        self.priority = priority
        self.repeatable = repeatable
        self.attr_props = attr_props
        self.inherit = inherit
        self.terminable = terminable


class Selector:
    def __init__(self,
                 expr: str,
                 selector: 'Selector' = None,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        if selector and not isinstance(selector, Selector):
            raise ValueError('selector must be a Selector')
        self.expr = expr
        self.selector = selector
        self.multi = multi
        self.trim = trim
        self.filter_empty = filter_empty
        self.format_str = format_str
        self.converter = converter

    def select(self, content: str, resp: Response) -> str | list[str]:
        try:
            selected = []
            if content and self.selector:
                content = self.selector.select(content, resp)
            if content and self.expr is not None:
                if isinstance(content, list):
                    for c in content:
                        selected.append(*self.do_select(c, resp))
                else:
                    selected = [*self.do_select(content, resp)]

            selected = selected if self.multi else ([selected[0]] if len(selected) > 0 else [])

            if selected and self.trim:
                selected = [x.strip() for x in selected]

            if self.filter_empty:
                selected = [x for x in selected if x]

            if self.format_str:
                selected = [self.format_str.format(x) for x in selected]

            if self.converter:
                if self.multi:
                    selected = [self.converter(x) for x in selected]
                else:
                    selected = [self.converter(selected[0]) if len(selected) > 0 else self.converter(None)]
            return selected if self.multi else (selected[0] if len(selected) > 0 else None)
        except BaseException as e:
            _logger.error(f"failed to select value from [{content} with selector [{self}]")
            raise e

    @abstractmethod
    def do_select(self, content: str, resp: Response) -> List[str]:
        pass

    def __str__(self):
        return f'{self.__class__.__name__}{{expr={self.expr}}}'

    __repr__ = __str__


class Embedded:
    def __init__(self, selector: Selector, embedded_class: type[R], *args, **kwargs):
        self.selector = selector
        self.target = embedded_class
        self.args = args
        self.kwargs = kwargs

    def select(self, content: str, resp: Response, links: list[Request] = None) -> R | list[R]:
        if links is None:
            links = []
        selected = self.selector.select(content, resp)
        if isinstance(selected, list):
            results = []
            for x in selected:
                s = select(x, resp, self.target, *self.args, **self.kwargs)
                results.append(s[0])
                links.extend(s[1])
            return results
        s = select(selected, resp, self.target, *self.args, **self.kwargs)
        links.extend(s[1])
        return s[0]


def embedded(selector: Selector, embedded_class: type[R], *args, **kwargs) -> Embedded:
    return Embedded(selector, embedded_class, *args, **kwargs)


def _get_type_selectors_links(cls: type) -> (dict[str, Selector | Embedded], List[Any]):
    selectors = {}
    links = []
    for t in [*cls.__bases__, cls]:
        for k, v in t.__dict__.items():
            if isinstance(v, Selector) or isinstance(v, Embedded):
                selectors[k] = v
        links.extend(_get_links(t))
    return selectors, links


def select(content: str, resp: Response, result_class: type, links: list[Request] = None, *args, **kwargs) -> (
        R, list[Request]):
    r = result_class(*args, **kwargs)
    _selectors, _links = _get_type_selectors_links(type(r))
    if links is None:
        links = []
    for key, value in _selectors.items():
        if isinstance(value, Selector):
            r.__dict__[key] = value.select(content, resp)
        elif isinstance(value, Embedded):
            r.__dict__[key] = value.select(content, resp, links)

    for link in _links:
        if link.terminable and link.terminable(r, content, resp):
            continue
        requests = []
        l = link.selector.select(content, resp)
        if l:
            if isinstance(l, list):
                requests.extend(l)
            else:
                requests.append(l)
            new_requests = [Request(x,
                                    link.method,
                                    queries=link.queries,
                                    data=link.data,
                                    headers=link.headers,
                                    priority=link.priority,
                                    repeatable=link.repeatable,
                                    inherit=link.inherit) for x in requests]
            if link.attr_props:
                for attr_prop in link.attr_props:
                    if attr_prop in r.__dict__:
                        for new_request in new_requests:
                            new_request.set_attr(attr_prop, r.__dict__[attr_prop])

            links.extend(new_requests)
    return r, links


def link(selector: Selector, method: str = 'GET',
         *,
         queries: dict[str, list[str]] = None,
         data: str = None,
         headers: dict[str, str] = None,
         priority: int = 0,
         repeatable: bool = True,
         attr_props: list[str] = None,
         inherit: bool = False,
         terminable: Terminable = None) -> Link:
    return Link(selector,
                method,
                queries=queries,
                data=data,
                headers=headers,
                priority=priority,
                repeatable=repeatable,
                attr_props=attr_props,
                inherit=inherit,
                terminable=terminable)


_LINKS: dict[type, list[Link]] = {}


def _get_links(cls: type) -> list[Link]:
    return _LINKS.get(cls, [])


def hyperlink(*links: Link):
    def decorator(cls):
        _LINKS[cls] = [*links]
        return cls

    return decorator



from .limiter import Limiter


class Site:
    def __init__(self,
                 host: str,
                 *,
                 limiter: Limiter = None,
                 headers: dict[str, str] = None,
                 proxy: str = None,
                 encoding: str = 'utf-8',
                 timeout: float = 30):
        self._host = host
        self._limiter = limiter
        self._headers = headers if headers is not None else {}
        self._proxy = proxy
        self._encoding = encoding
        self._timeout = timeout

    @property
    def host(self):
        return self._host

    @property
    def limiter(self):
        return self._limiter

    @property
    def headers(self):
        return self._headers

    @property
    def proxy(self):
        return self._proxy

    @property
    def encoding(self):
        return self._encoding

    @property
    def timeout(self):
        return self._timeout

    def __str__(self):
        return f'{{host={self.host}}}'


def new(host: str,
        *,
        limiter: Limiter = None,
        headers: dict[str, str] = None,
        proxy: str = None,
        encoding: str = 'utf-8',
        timeout: float = 30) -> Site:
    return Site(host, limiter=limiter, headers=headers, proxy=proxy, encoding=encoding, timeout=timeout)



import re
from urllib.parse import urlparse

from ..types import Matcher


def and_matcher(*matchers: Matcher) -> Matcher:
    return lambda res: all(m(res) for m in matchers)


def or_matcher(*matchers: Matcher) -> Matcher:
    return lambda res: any(m(res) for m in matchers)


def not_matcher(matcher: Matcher) -> Matcher:
    return lambda res: not matcher(res)


def host_matcher(host: str) -> Matcher:
    return lambda res: urlparse(res.request.url).hostname == host


def url_matcher(regex: str) -> Matcher:
    r = re.compile(regex)
    return lambda res: bool(r.match(res.request.url))


def content_type_matcher(regex: str) -> Matcher:
    return header_matcher('content-type', regex)


def header_matcher(header: str, regex: str) -> Matcher:
    r = re.compile(regex)
    return lambda res: bool(r.match(res.headers.get(header.lower(), '')))


ALL: Matcher = lambda res: True

JSON: Matcher = content_type_matcher(r'.*application/json.*')

HTML: Matcher = content_type_matcher(r'.*text/html.*')

IMAGE: Matcher = content_type_matcher(r'.*image.*')

VIDEO: Matcher = content_type_matcher(r'.*video.*')

PDF: Matcher = content_type_matcher(r'.*application/pdf.*')

WORD: Matcher = content_type_matcher(r'.*application/msword.*')

EXCEL: Matcher = content_type_matcher(r'.*application/vnd.ms-excel.*')

AUDIO: Matcher = content_type_matcher(r'.*audio.*')

OCTET_STREAM = content_type_matcher(r'.*application/octet-stream.*')

MEDIA = or_matcher(IMAGE, VIDEO, AUDIO)



import threading
from datetime import datetime, timedelta
from time import sleep


class Limiter:
    def __init__(self, interval_in_seconds: float, capacity: int):
        self._capacity = capacity
        self._interval = interval_in_seconds
        self._count = 0
        self._last_time = datetime.now()
        self._lock = threading.RLock()

    def acquire(self, timeout_milliseconds: int = None) -> bool:
        with self._lock:
            if timeout_milliseconds is None or timeout_milliseconds <= 0:
                return self._acquire()
            else:
                return self._acquire(datetime.now() + timedelta(milliseconds=timeout_milliseconds))

    def _acquire(self, end_time: datetime = None) -> bool:
        now = datetime.now()
        self._count = min(self._capacity,
                          self._count + int((now.timestamp() - self._last_time.timestamp()) / self._interval))
        if self._count > 0:
            self._count = self._count - 1
            self._last_time = now
            return True
        if end_time is None:
            sleep(self._interval - now.timestamp() + self._last_time.timestamp())
            return self._acquire(end_time)
        else:
            if now > end_time:
                return False
            sleep(min(self._interval, self._interval - now.timestamp() + self._last_time.timestamp()))
            return self._acquire(end_time)


def new(interval_in_seconds: float = 1, capacity: int = 1) -> Limiter:
    return Limiter(interval_in_seconds, capacity)



from abc import abstractmethod

from ..request import Request, State


class Store:

    @abstractmethod
    def put(self, r: Request) -> bool:
        pass

    @abstractmethod
    def get(self) -> Request | None:
        pass

    @abstractmethod
    def exists(self, id: str) -> bool:
        pass

    @abstractmethod
    def update_state(self, r: Request, state: State, msg: str = None):
        pass

    @abstractmethod
    def reply_failed(self) -> int:
        pass

    @abstractmethod
    def get_statistics(self) -> tuple[int, int, int, int, int]:
        pass

    @abstractmethod
    def has_waiting_requests(self) -> bool:
        pass



from .memory_store import new as memory_store
from .redis_store import new as redis_store
from .sqlite_store import new as sqlite_store
from .store import Store

__all__ = ['memory_store', 'sqlite_store', 'redis_store', 'Store']



import json
from enum import Enum
from typing import Any


class State(Enum):
    NEW = 'NEW'
    WAITING = 'WAITING'
    EXECUTING = 'EXECUTING'
    FAILED = 'FAILED'
    COMPLETED = 'COMPLETED'


class Request:
    def __init__(self, url: str, method: str = 'GET',
                 *,
                 queries: dict[str, list[Any] | Any] = None,
                 data: str = None,
                 headers: dict[str, str] = None,
                 priority: int = 0,
                 repeatable: bool = True,
                 attrs: dict[str, Any] = None,
                 inherit: bool = False):
        if not url:
            raise ValueError('url is empty')
        if not method:
            raise ValueError('method is empty')
        if method not in ('GET', 'POST'):
            raise ValueError('method can only be GET or POST')
        self.url = url
        self.method = method
        self.queries = {}
        if queries:
            for k, v in queries.items():
                if not isinstance(v, list):
                    self.queries[k] = [str(v)]
                else:
                    self.queries[k] = v
        self.data = data
        self.headers = {} if headers is None else headers
        self.priority = priority
        self.repeatable = repeatable
        self.attrs = {} if attrs is None else attrs
        self.inherit = inherit
        self.parent = None
        self.id = None
        self.state = State.NEW
        self.msg = None
        self.depth = 1

    def get_attr(self, name):
        return self.attrs.get(name, None)

    def set_attr(self, name, value):
        self.attrs[name] = value

    def __lt__(self, other):
        return self.priority < other.priority

    def __str__(self):
        return f'{{id={self.id}, url={self.url}, priority={self.priority}, depth={self.depth}}}'

    def to_json(self) -> str:
        return json.dumps({
            'id': self.id,
            'url': self.url,
            'method': self.method,
            'queries': self.queries,
            'data': self.data,
            'headers': self.headers,
            'priority': self.priority,
            'repeatable': self.repeatable,
            'attrs': self.attrs,
            'inherit': self.inherit,
            'parent': self.parent,
            'state': self.state.value,
            'msg': self.msg,
            'depth': self.depth
        })

    @staticmethod
    def from_json(json_str: str) -> 'Request':
        json_object = json.loads(json_str)
        req = Request(json_object['url'])
        req.id = json_object['id']
        req.method = json_object['method']
        req.queries = json_object['queries']
        req.data = json_object['data']
        req.headers = json_object['headers']
        req.priority = json_object['priority']
        req.repeatable = json_object['repeatable']
        req.attrs = json_object['attrs']
        req.inherit = json_object['inherit']
        req.parent = json_object['parent']
        req.state = State(json_object['state'])
        req.msg = json_object['msg']
        req.depth = json_object['depth']
        return req

    __repr__ = __str__


def new(url: str, method: str = 'GET',
        *,
        queries: dict[str, list[Any] | Any] = None,
        data: str = None,
        headers: dict[str, str] = None,
        priority: int = 0,
        repeatable: bool = True,
        attrs: dict[str, Any] = None,
        inherit: bool = False) -> Request:
    return Request(url, method, queries=queries, data=data, headers=headers, priority=priority, repeatable=repeatable,
                   attrs=attrs, inherit=inherit)



import queue

from .store import Store
from ..request import Request, State


class _Wrapper:
    def __init__(self, id: str, priority: int):
        self.id = id
        self.priority = priority

    def __lt__(self, other):
        return not (self.priority < other.priority)


class _MemoryStore(Store):
    def __init__(self):
        self._queue = queue.PriorityQueue()
        self._all: dict[str, Request] = {}
        self._fails = set()
        self._executing = set()
        self._completed = set()

    def put(self, r: Request) -> bool:
        self._all[r.id] = r
        self._queue.put(_Wrapper(r.id, r.priority))
        return True

    def get(self) -> Request | None:
        try:
            r = self._queue.get(False)
            req = self._all[r.id] if r else None
            if req:
                req.state = State.EXECUTING
                req.msg = "正在处理"
                self._executing.add(r.id)
            return req
        except queue.Empty:
            return None

    def update_state(self, r: Request, state: State, msg: str = None):
        if state == State.COMPLETED:
            self._completed.add(r.id)
            self._fails.discard(r.id)
            self._executing.discard(r.id)
        elif state == State.FAILED:
            self._fails.add(r.id)
            self._completed.discard(r.id)
            self._executing.discard(r.id)
        elif state == State.WAITING:
            self._completed.discard(r.id)
            self._executing.discard(r.id)
            self._fails.discard(r.id)
            self._queue.put(_Wrapper(r.id, r.priority))
        else:
            raise ValueError(f"Invalid state: {state}")

    def exists(self, id: str) -> bool:
        return id in self._all

    def reply_failed(self) -> int:
        fails = [self._all[f] for f in self._fails]
        for fail in fails:
            fail.state = State.WAITING
            fail.msg = "等待处理"
            self.update_state(fail, State.WAITING, "等待处理")
        return len(fails)

    def get_statistics(self) -> tuple[int, int, int, int, int]:
        return len(self._all), self._queue.qsize(), len(self._executing), len(self._completed), len(self._fails)

    def has_waiting_requests(self) -> bool:
        return self._queue.qsize() > 0 or len(self._executing) > 0


def new() -> Store:
    return _MemoryStore()



import hashlib
import logging
import os
import queue
import re
import threading
from concurrent.futures import ThreadPoolExecutor, wait, Future, ALL_COMPLETED
from enum import Enum
from urllib.parse import urljoin, urlencode, parse_qs, urlparse

import curl_cffi.requests as curl_cffi
import requests

from .downloader import requests_downloader

from .request import Request, State as RequestState
from .response import Response
from .site import Site
from .store import Store, memory_store
from .types import Processor, Matcher, Downloader

_HEADER_COOKIE = "Cookie"
_HEADER_REFERER = "Referer"
_HEADER_UA = "User-Agent"
_DEFAULT_HEADERS = {
    _HEADER_UA: "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
}

_REGEX_REFERER = re.compile(r"^(https?://([^/]+)).*$")

_logger = logging.getLogger("pyoctopus")


def _generate_request_id(r: Request) -> str:
    parsed_url = urlparse(r.url)
    existing_params = parse_qs(parsed_url.query)
    if r.queries:
        for k, v in r.queries.items():
            q = []
            q.extend(existing_params.get(k, []))
            q.extend(v if v else [])
            q = sorted(q)
            existing_params[k] = q
    url = parsed_url._replace(query=urlencode(sorted(existing_params.items()), doseq=True)).geturl()
    if r.data:
        url = f"{url}&{r.data}"
    url = f"{r.method}{url}"
    md5 = hashlib.md5()
    md5.update(url.encode("utf-8"))
    return md5.hexdigest()


class State(Enum):
    INIT = 0
    STARTING = 1
    STARTED = 2
    STOPPING = 3
    STOPPED = 4


class Octopus:
    def __init__(
        self,
        downloader: Downloader = None,
        store: Store = None,
        processors: list[tuple[Matcher, Processor]] = None,
        threads: int = os.cpu_count(),
        queue_factor: int = 2,
        sites: list[Site] = None,
        retries: int = 1,
        ignore_seed_when_has_waiting_requests: bool = False,
    ):
        self.downloader = downloader or requests_downloader
        self._store = store or memory_store()
        self._seeds = []
        self._processors = processors if processors is not None else []
        self._threads = threads
        self._queue_factor = queue_factor
        if sites:
            self._sites = {s.host: s for s in sites}
        self.retries = retries
        self._ignore_seed_when_has_waiting_requests = ignore_seed_when_has_waiting_requests
        self._lock = threading.Lock()
        self._semaphore = threading.Semaphore(self._queue_factor * self._threads)
        self._workers = None
        self._workers_futures = []
        self._boss = None
        self._boss_future = None
        self._queue = queue.Queue()
        self._state = State.INIT

    def start_async(self, *seeds: Request | str) -> Future[None]:
        if not self._set_state(State.STARTING, State.INIT):
            raise RuntimeError("Pyoctopus is not in INIT state")
        self._state = State.STARTING
        if not self._ignore_seed_when_has_waiting_requests or not self._store.has_waiting_requests():
            self._seeds.extend([Request(s) if isinstance(s, str) else s for s in seeds])
            for seed in self._seeds:
                self._add(seed)
        self._boss = ThreadPoolExecutor(max_workers=1, thread_name_prefix="boss")
        self._workers = ThreadPoolExecutor(max_workers=self._threads, thread_name_prefix="worker")
        self._boss_future = self._boss.submit(self._dispatch)
        _logger.info("Pyoctopus started")
        self._state = State.STARTED
        return self._boss_future

    def start(self, *seeds: Request | str):
        self.start_async(*seeds).result()

    def stop(self):
        if not self._set_state(State.STOPPING, State.STARTED):
            raise RuntimeError("Pyoctopus is not in STARTED state")
        wait([self._boss_future, *self._workers_futures], return_when=ALL_COMPLETED)
        self._boss.shutdown()
        self._workers.shutdown()
        self._state = State.STOPPED
        stat = self._store.get_statistics()
        _logger.info(
            f"Pyoctopus stats: all = {stat[0]}, waiting = {stat[1]}, executing = {stat[2]}, completed = {stat[3]}, failed = {stat[4]}"
        )
        _logger.info("Pyoctopus stopped")

    def add(self, r: Request) -> None:
        with self._lock:
            if self._state.value > State.STARTED.value:
                raise RuntimeError(f"Pyoctopus is in {self._state} state")
        self._add(r)

    def _add(self, r: Request, p: Request = None) -> None:
        if p is not None:
            r.parent = p.id
            r.depth = p.depth + 1
            if r.inherit:
                r.headers = {**p.headers, **r.headers}
                r.attrs = {**p.attrs, **r.attrs}
            if r.headers.get(_HEADER_REFERER, None) is None:
                m = _REGEX_REFERER.match(p.url)
                if m is not None:
                    r.headers[_HEADER_REFERER] = m.group(1)
            if not r.url.startswith("http"):
                r.url = urljoin(p.url, r.url)
        r.id = _generate_request_id(r)
        r.state = RequestState.WAITING
        r.msg = "等待处理"

        def _r():
            if r.repeatable or not self._store.exists(r.id):
                if not self._store.put(r):
                    _logger.warning(f"Can not put [{r}] to store")

        self._queue.put(_r)

    @property
    def state(self):
        with self._lock:
            return self._state

    def _dispatch(self) -> None:
        while True:
            has_queued_tasks = False
            while True:
                try:
                    r = self._queue.get(False)
                    if r is not None:
                        has_queued_tasks = True
                        r()
                    else:
                        break
                except queue.Empty:
                    break
            self._workers_futures = [f for f in self._workers_futures if not f.done()]

            with self._lock:
                if self._state.value >= State.STOPPING.value:
                    if not has_queued_tasks and len(self._workers_futures) == 0:
                        break
                else:
                    r = self._store.get()
                    if r is not None:
                        _logger.info(f"Take {r}")
                        self._semaphore.acquire()
                        self._workers_futures.append(self._workers.submit(self._process, r))
                    if r is None and len(self._workers_futures) == 0 and not has_queued_tasks:
                        if not self._retry_fails():
                            _logger.info("No more tasks found, pyoctopus will stop")
                            threading.Thread(target=self.stop, name="StopThread").start()
                            break
        if self._state.value > State.STARTED.value:
            self._log_undone_tasks()

    def _retry_fails(self) -> bool:
        has_fails = False
        if self.retries > 0:
            count = self._store.reply_failed()
            if count > 0:
                has_fails = True
                _logger.info(f"[{self.retries}] Retry {count} failed requests")
            self.retries = self.retries - 1
        return has_fails

    def _process(self, r: Request):
        res = None
        try:
            site = self._get_site(urlparse(r.url).hostname)
            if site.limiter is not None:
                site.limiter.acquire()
            res = self._download(r, site)
            if res.status != 200:
                raise ValueError(f"Bad http status [{res.status}] for [{r}]")
            for [m, p] in self._processors:
                if m and m(res):
                    for req in p(res):
                        self._add(req, r)
            r.msg = "成功处理"
            r.state = RequestState.COMPLETED
            self._queue.put(lambda: self._store.update_state(r, RequestState.COMPLETED, "成功处理"))
        except BaseException as e:
            r.msg = str(e)
            r.state = RequestState.FAILED
            self._queue.put(lambda: self._store.update_state(r, RequestState.FAILED, r.msg))
            _logger.error(f"Process [req = {r}, resp = {res}] error\n{r.msg}", exc_info=True)
        finally:
            self._semaphore.release()

        if self._state.value > State.STARTED.value:
            self._log_undone_tasks()

    def _set_state(self, new_state: State, expected_state: State) -> bool:
        with self._lock:
            if self._state == expected_state:
                self._state = new_state
                _logger.debug(f"Pyoctopus state changed from {expected_state} to {new_state}")
                return True
            else:
                return False

    def _download(self, request: Request, site: Site) -> Response:
        try:
            return self.downloader(request, site)
        except BaseException as e:
            raise RuntimeError(str(e))

    def _get_site(self, host: str) -> Site:
        if host in self._sites:
            return self._sites[host]
        for h, s in self._sites.items():
            if re.match(h.replace("*", ".*"), host):
                return s
        return Site(host)

    def _log_undone_tasks(self):
        undone_count = len([x for x in [self._boss_future, *self._workers_futures] if not x.done()])
        if undone_count > 0:
            _logger.info(f"Wait for {undone_count} tasks in the queue to complete")


def new(
    downloader: Downloader = None,
    store: Store = None,
    processors: list[tuple[Matcher, Processor]] = None,
    threads: int = os.cpu_count(),
    queue_factor: int = 2,
    sites: list[Site] = None,
    retries: int = 1,
    ignore_seed_when_has_waiting_requests: bool = False,
) -> Octopus:
    return Octopus(
        downloader=downloader,
        store=store,
        processors=processors,
        threads=threads,
        queue_factor=queue_factor,
        sites=sites,
        retries=retries,
        ignore_seed_when_has_waiting_requests=ignore_seed_when_has_waiting_requests,
    )



from .selector import Selector
from .. import Response
from ..types import Converter


class Attr(Selector):
    def __init__(self,
                 expr: str,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Attr, self).__init__(expr,
                                   multi=multi,
                                   trim=trim,
                                   filter_empty=filter_empty,
                                   format_str=format_str,
                                   converter=converter)

    def do_select(self, content: str, resp: Response) -> list[str]:
        attr = resp.request.get_attr(self.expr)
        if attr is None:
            return []
        return [str(c) for c in attr] if isinstance(attr, list) else [str(attr)]


def new(name: str,
        *,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Attr:
    return Attr(name,
                multi=multi,
                trim=trim,
                filter_empty=filter_empty,
                format_str=format_str,
                converter=converter)



import re

import redis

from .store import Store
from ..request import State, Request


class RedisStore(Store):

    def __init__(
        self, *, prefix: str = "pyoctopus", host: str = "127.0.0.1", port: int = 6379, db: int = 0, password: str = None
    ):
        if ":" in prefix:
            raise ValueError("Prefix cannot contain colon")
        self._prefix = prefix
        self._pool = redis.ConnectionPool(host=host, port=port, db=db, password=password)
        self._client = redis.Redis(connection_pool=self._pool)
        self._waiting_priority_pattern = re.compile(f"^{prefix}:waiting:(.*):(.*)$")
        # executing -> waiting
        self._re_waiting()

    def put(self, r: Request) -> bool:
        self._client.set(f"{self._prefix}:all:{r.id}", r.to_json())
        self._client.set(f"{self._prefix}:waiting:{r.priority}:{r.id}", "")
        return True

    def get(self) -> Request | None:
        cursor = 0
        while True:
            cursor, keys = self._client.scan(cursor, match=f"{self._prefix}:waiting:*", count=20)
            if keys:
                key = sorted([k.decode() for k in keys], reverse=True)[0]
                m = self._waiting_priority_pattern.match(key)
                id = m.group(2)
                req = Request.from_json(self._client.get(f"{self._prefix}:all:{id}"))
                self._client.delete(key)
                self._client.set(f"{self._prefix}:executing:{id}", req.priority)
                return req
            if cursor == 0:
                return None

    def exists(self, id: str) -> bool:
        return self._client.exists(f"{self._prefix}:all:{id}")

    def update_state(self, r: Request, state: State, msg: str = None):
        self._client.set(f"{self._prefix}:all:{r.id}", r.to_json())
        if state == State.COMPLETED:
            self._client.set(f"{self._prefix}:completed:{r.id}", "")
            self._client.delete(f"{self._prefix}:executing:{r.id}")
            self._client.delete(f"{self._prefix}:failed:{r.id}")
        elif state == State.FAILED:
            self._client.set(f"{self._prefix}:failed:{r.id}", r.priority)
            self._client.delete(f"{self._prefix}:executing:{r.id}")
            self._client.delete(f"{self._prefix}:completed:{r.id}")
        elif state == State.WAITING:
            self._client.delete(f"{self._prefix}:failed:{r.id}")
            self._client.delete(f"{self._prefix}:executing:{r.id}")
            self._client.delete(f"{self._prefix}:completed:{r.id}")
            self._client.set(f"{self._prefix}:waiting:{r.priority}:{r.id}", "")
        else:
            raise ValueError(f"Invalid state: {state}")

    def reply_failed(self) -> int:
        cursor = 0
        count = 0
        while True:
            cursor, keys = self._client.scan(cursor, match=f"{self._prefix}:failed:*", count=100)
            count += len(keys)
            for k in keys:
                self._client.set(f'{self._prefix}:waiting:{self._client.get(k)}:{k.split(":")[-1]}', "")
                self._client.delete(k)
            if cursor == 0:
                break
        return count

    def get_statistics(self) -> tuple[int, int, int, int, int]:
        return (
            self._get_key_size(f"{self._prefix}:all:*"),
            self._get_key_size(f"{self._prefix}:waiting:*"),
            self._get_key_size(f"{self._prefix}:executing:*"),
            self._get_key_size(f"{self._prefix}:completed:*"),
            self._get_key_size(f"{self._prefix}:failed:*"),
        )

    def _get_key_size(self, key: str) -> int:
        cursor = 0
        count = 0
        while True:
            cursor, keys = self._client.scan(cursor, match=key, count=100)
            count += len(keys)
            if cursor == 0:
                break
        return count

    def _re_waiting(self):
        cursor = 0
        while True:
            cursor, keys = self._client.scan(cursor, match=f"{self._prefix}:executing:*", count=100)
            for k in keys:
                priority = self._client.get(k)
                self._client.set(f'{self._prefix}:waiting:{priority}:{k.split(":")[-1]}', "")
                self._client.delete(k)
            if cursor == 0:
                break

    def has_waiting_requests(self) -> bool:
        return self._client.exists(f"{self._prefix}:waiting:*") or self._client.exists(f"{self._prefix}:executing:*")


def new(
    *, prefix: str = "pyoctopus", host: str = "127.0.0.1", port: int = 6379, db: int = 0, password: str = None
) -> RedisStore:
    return RedisStore(prefix=prefix, host=host, port=port, db=db, password=password)



import re

from .selector import Selector
from .. import Response
from ..types import Converter


class Regex(Selector):
    def __init__(self,
                 expr: str,
                 group: int | list[int] = 0,
                 selector: Selector = None,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Regex, self).__init__(expr,
                                    selector=selector,
                                    multi=multi,
                                    trim=trim,
                                    filter_empty=filter_empty,
                                    format_str=format_str,
                                    converter=converter)
        self.group = group if isinstance(group, list) else [group]

    def do_select(self, content: str, resp: Response) -> list[str]:
        matches = re.finditer(self.expr, content)
        return [(''.join([x.group(g) for g in self.group])) for x in matches]


def new(expr: str,
        group: int | list[int] = 0,
        selector: Selector = None,
        *,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Regex:
    return Regex(expr,
                 group,
                 selector,
                 multi=multi,
                 trim=trim,
                 filter_empty=filter_empty,
                 format_str=format_str,
                 converter=converter)



from enum import Enum
from typing import TypeVar, Callable, Any

from .site import Site

from .request import Request
from .response import Response

R = TypeVar("R", bound=[type])

Converter = TypeVar("Converter", bound=[Callable[[str], Any]])

Matcher = TypeVar("Matcher", bound=[Callable[[Response], bool]])

Downloader = TypeVar("Downloader", bound=[Callable[[Request, Site], Response]])

Processor = TypeVar("Processor", bound=[Callable[[Response], list[Request]]])

Collector = TypeVar("Collector", bound=[Callable[[R], None]])

Terminable = TypeVar("Terminable", bound=[Callable[[R, str, Response], bool]])



import os.path

import openpyxl
from openpyxl.styles import Font, PatternFill, Alignment, Border, Side

from ..types import Collector, R


class CellStyle:
    def __init__(self, *,
                 font_size: int = 11,
                 font_bold=False,
                 font_color: str = '000000',
                 background_color: str = 'FFFFFF',
                 border_color: str = 'd4d4d4',
                 alignment: str = 'left',
                 delimiter: str = '\n'):
        self.font_size = font_size
        self.font_bold = font_bold
        self.font_color = font_color
        self.background_color = background_color
        self.border_color = border_color
        self.alignment = alignment
        self.delimiter = delimiter


class Column:
    def __init__(self, field: str, name: str, style: CellStyle = None):
        self.field = field
        self.name = name
        if style is None:
            self.style = CellStyle()
        else:
            self.style = style


def _set_cell_style(cell, style: CellStyle):
    cell.font = Font(size=style.font_size, bold=style.font_bold, color=style.font_color)
    cell.fill = PatternFill(fill_type="solid", start_color=style.background_color,
                            end_color=style.background_color)
    cell.alignment = Alignment(horizontal=style.alignment, vertical="center", wrap_text=True)
    cell.border = Border(
        left=Side(style="thin", color=style.border_color),
        right=Side(style="thin", color=style.border_color),
        top=Side(style="thin", color=style.border_color),
        bottom=Side(style="thin", color=style.border_color),
    )


def new(file: str, append: bool = False, header_style: CellStyle = None,
        columns: list[Column] = None) -> Collector:
    if header_style is None:
        header_style = CellStyle(font_size=12, font_bold=True, font_color='FFFFFF', background_color='808080',
                                 alignment='left')

    if not append and os.path.exists(file):
        os.remove(file)

    if os.path.exists(file):
        wb = openpyxl.load_workbook(file)
    else:
        wb = openpyxl.Workbook()

    sheet = wb.active

    if columns and sheet.max_row == 1:
        sheet.append([c.name for c in columns])
        for cell in sheet[1]:
            _set_cell_style(cell, header_style)
        wb.save(file)

    def _collect(r: R) -> None:
        keys = r.__dict__.keys()
        _vals = []
        if not columns:
            values = [(r.__dict__.get(k, None), None) for k in keys]
        else:
            values = [(r.__dict__.get(c.field, None), c) for c in columns]
        for value in values:
            if value[0] is None:
                _vals.append('')
            elif isinstance(value[0], list):
                _vals.append((value[1].style.delimiter if value[1] else '\n').join([str(v) for v in value[0]]))
            else:
                _vals.append(str(value[0]))
        sheet.append(_vals)
        for i, cell in enumerate(sheet[sheet.max_row]):
            col_style = columns[i].style if columns and columns[i].style else CellStyle()
            _set_cell_style(cell, col_style)

        # 自动设置列宽
        for column in sheet.columns:
            max_length = 0
            column_letter = column[0].column_letter  # 获取列字母
            for cell in column:
                length = sum(
                    2 if '\u4e00' <= char <= '\u9fff' else 1 for char in str(cell.value))
                max_length = max(max_length, length)
            adjusted_width = max_length + 2
            sheet.column_dimensions[column_letter].width = adjusted_width

        wb.save(file)

    return _collect


def new_column(field: str, column: str, style: CellStyle = None) -> Column:
    return Column(field, column, style)


def new_cell_style(*,
                   font_size: int = 12,
                   font_bold=False,
                   font_color: str = '000000',
                   background_color: str = 'FFFFFF',
                   border_color: str = 'd4d4d4',
                   alignment: str = 'left',
                   delimiter: str = '\n') -> CellStyle:
    return CellStyle(font_size=font_size, font_bold=font_bold, font_color=font_color, border_color=border_color,
                     background_color=background_color, alignment=alignment, delimiter=delimiter)



import urllib
from urllib.parse import urlencode

from .selector import Selector
from .. import Response
from ..types import Converter


class Url(Selector):
    def __init__(self,
                 url_decode: bool = False,
                 url_encode: bool = False,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Url, self).__init__('',
                                  multi=multi,
                                  trim=trim,
                                  filter_empty=filter_empty,
                                  format_str=format_str,
                                  converter=converter)
        self.encode = url_encode
        self.decode = url_decode

    def do_select(self, content: str, resp: Response) -> list[str]:
        r = resp.request.url
        if self.encode:
            r = urllib.parse.quote(r)
        if self.decode:
            r = urllib.parse.unquote(r)
        return [r]


def new(url_decode: bool = False,
        url_encode: bool = False,
        *,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Url:
    return Url(
        url_decode=url_decode,
        url_encode=url_encode,
        multi=multi,
        trim=trim,
        filter_empty=filter_empty,
        format_str=format_str,
        converter=converter)



import json
import sqlite3
import threading

from .store import Store
from ..request import Request, State

_COL_ID = ("id", "TEXT PRIMARY KEY")
_COLS = [
    ("url", "TEXT"),
    ("method", "TEXT"),
    ("priority", "INTEGER"),
    ("repeatable", "INTEGER"),
    ("parent", "TEXT"),
    ("data", "BLOB"),
    ("queries", "TEXT"),
    ("headers", "TEXT"),
    ("attrs", "TEXT"),
    ("state", "TEXT"),
    ("depth", "INTEGER"),
    ("msg", "TEXT"),
    ("inherit", "INTEGER"),
]

_SQL_CREATE_TABLE = "CREATE TABLE IF NOT EXISTS {} (" + ", ".join([f"{c[0]} {c[1]}" for c in [_COL_ID, *_COLS]]) + ")"

_COL_NAMES = ", ".join([c[0] for c in [_COL_ID, *_COLS]])

_COL_UPDATE_BY_ID = ", ".join([f"{c[0]} = ?" for c in [*_COLS]])

_local = threading.local()


class SqliteStore(Store):
    def __init__(self, db: str, table: str = "pyoctopus"):
        super(SqliteStore, self).__init__()
        self._db = db
        self._table = table

        self._sql_create_table = _SQL_CREATE_TABLE.format(self._table)
        self._sql_create_idx_priority = (
            f"CREATE INDEX IF NOT EXISTS idx_{self._table}_priority on {self._table}(priority)"
        )
        self._sql_exist_by_id = f"SELECT count(1) FROM {self._table} WHERE id = ?"
        self._sql_update_state_by_id = f"UPDATE {self._table} SET state = ?, msg = ? WHERE id = ?"
        self._sql_get = f"SELECT {_COL_NAMES} FROM {self._table} WHERE state = ? ORDER BY priority DESC LIMIT 1"
        self._sql_update_by_id = f"UPDATE {self._table} SET {_COL_UPDATE_BY_ID} WHERE id = ?"
        self._sql_put = (
            f'INSERT INTO {self._table} ({_COL_NAMES}) VALUES ({", ".join(["?" for _ in [_COL_ID, *_COLS]])})'
        )
        self._sql_update_state = f"UPDATE {self._table} SET state = ?, msg = ? WHERE state = ?"
        self._sql_paged_select = f"SELECT {_COL_NAMES} FROM {self._table} WHERE state = ? LIMIT ? OFFSET ?"
        self._sql_count_by_state = f"SELECT count(1) FROM {self._table} WHERE state = ?"
        self._init_table()

    def put(self, r: Request) -> bool:
        with self._get_connection() as _connection:
            try:
                _cursor = _connection.cursor()
                if self.exists(r.id):
                    _cursor.execute(
                        self._sql_update_by_id,
                        (
                            r.url,
                            r.method,
                            r.priority,
                            r.repeatable,
                            r.parent,
                            r.data,
                            json.dumps(r.queries, ensure_ascii=False),
                            json.dumps(r.headers, ensure_ascii=False),
                            json.dumps(r.attrs, ensure_ascii=False),
                            State.WAITING.value,
                            r.depth,
                            r.msg,
                            r.inherit,
                            r.id,
                        ),
                    )
                else:
                    _cursor.execute(
                        self._sql_put,
                        (
                            r.id,
                            r.url,
                            r.method,
                            r.priority,
                            r.repeatable,
                            r.parent,
                            r.data,
                            json.dumps(r.queries, ensure_ascii=False),
                            json.dumps(r.headers, ensure_ascii=False),
                            json.dumps(r.attrs, ensure_ascii=False),
                            r.state.value,
                            r.depth,
                            r.msg,
                            r.inherit,
                        ),
                    )
                _connection.commit()
                return True
            except sqlite3.Error as e:
                _connection.rollback()
                raise e

    def _row_to_request(self, row) -> Request:
        r = Request(
            row[1],
            method=row[2],
            priority=row[3],
            repeatable=row[4],
            data=row[6],
            queries=json.loads(row[7]),
            headers=json.loads(row[8]),
            attrs=json.loads(row[9]),
        )
        r.id = row[0]
        r.parent = row[5]
        r.state = State(row[10])
        r.depth = row[11]
        r.msg = row[12]
        r.inherit = bool(row[13])
        return r

    def get(self) -> Request | None:
        with self._get_connection() as _connection:
            try:
                _cursor = _connection.cursor()
                _cursor.execute(self._sql_get, (State.WAITING.value,))
                row = _cursor.fetchone()
                if row is not None:
                    r = self._row_to_request(row)
                    r.state = State.EXECUTING
                    r.msg = "正在处理"
                    _cursor.execute(
                        self._sql_update_state_by_id,
                        (
                            State.EXECUTING.value,
                            "正在处理",
                            r.id,
                        ),
                    )
                    return r
                _connection.commit()
                return None
            except sqlite3.Error as e:
                _connection.rollback()
                raise e

    def update_state(self, r: Request, state: State, msg: str = None):
        r.state = state
        r.msg = msg
        with self._get_connection() as _connection:
            try:
                _cursor = _connection.cursor()
                _cursor.execute(
                    self._sql_update_state_by_id,
                    (
                        state.value,
                        msg,
                        r.id,
                    ),
                )
                _connection.commit()
            except sqlite3.Error as e:
                _connection.rollback()
                raise e

    def exists(self, id: str) -> bool:
        with self._get_connection() as _connection:
            try:
                _cursor = _connection.cursor()
                _cursor.execute(self._sql_exist_by_id, (id,))
                row = _cursor.fetchone()
                _connection.commit()
                return row[0] > 0
            except sqlite3.Error as e:
                raise e

    def _init_table(self):
        with sqlite3.connect(self._db) as _connection:
            try:
                _cursor = _connection.cursor()
                _cursor.execute(self._sql_create_table.format(self._table))
                _cursor.execute(self._sql_create_idx_priority)
                _cursor.execute(self._sql_update_state, (State.WAITING.value, "等待处理", State.EXECUTING.value))
                _connection.commit()
            except sqlite3.Error as e:
                _connection.rollback()
                raise e

    def reply_failed(self) -> int:
        with self._get_connection() as _connection:
            try:
                _cursor = _connection.cursor()
                _cursor.execute(self._sql_update_state, (State.WAITING.value, "等待处理", State.FAILED.value))
                _connection.commit()
                return _cursor.rowcount
            except sqlite3.Error as e:
                _connection.rollback()
                raise e

    def _get_connection(self) -> sqlite3.Connection:
        if not hasattr(_local, "conn"):
            _local.conn = sqlite3.connect(self._db)
        return _local.conn

    def get_statistics(self) -> tuple[int, int, int, int, int]:
        with self._get_connection() as _connection:
            try:
                _cursor = _connection.cursor()
                _cursor.execute(f"select count(1) from {self._table}")
                all = _cursor.fetchone()[0]
                _cursor.execute(self._sql_count_by_state, (State.WAITING.value,))
                waiting = _cursor.fetchone()[0]
                _cursor.execute(self._sql_count_by_state, (State.EXECUTING.value,))
                executing = _cursor.fetchone()[0]
                _cursor.execute(self._sql_count_by_state, (State.COMPLETED.value,))
                completed = _cursor.fetchone()[0]
                _cursor.execute(self._sql_count_by_state, (State.FAILED.value,))
                failed = _cursor.fetchone()[0]
                _connection.commit()
                return all, waiting, executing, completed, failed
            except sqlite3.Error as e:
                _connection.rollback()
                raise e

    def has_waiting_requests(self) -> bool:
        with self._get_connection() as _connection:
            _cursor = _connection.cursor()
            _cursor.execute(self._sql_count_by_state, (State.WAITING.value,))
            if _cursor.fetchone()[0] <= 0:
                _cursor.execute(self._sql_count_by_state, (State.EXECUTING.value,))
                return _cursor.fetchone()[0] > 0
            return True


def new(db: str, table: str = "pyoctopus") -> SqliteStore:
    return SqliteStore(db, table)



from .site import new as site

from .request import Request
from .request import new as request
from .response import Response
from .response import new as response
from .limiter import new as limiter
from .octopus import new
from .types import R, Converter, Collector, Processor, Matcher, Terminable, Downloader

from .converter import *
from .selector import *
from .matcher import *
from .processor import *
from .store import *
from .collector import *

from .downloader import *



from .request import Request


class Response:
    def __init__(self,
                 request: Request,
                 *,
                 status: int = 0,
                 content: bytes = None,
                 headers: dict[str, str] = None,
                 encoding: str = 'utf-8'
                 ):
        self.request = request
        self.content = content
        self.status = status
        self.content = content
        self.headers = headers
        self.encoding = encoding
        self._text = None
        self._parsed = False

    @property
    def text(self):
        if self._parsed:
            return self._text
        self._text = self.content.decode(self.encoding)
        self._parsed = True
        return self._text

    def __str__(self):
        return f'{{request={self.request}, status={self.status}, length={len(self.content)}, encoding={self.encoding}}}'

    __repr__ = __str__


def new(request: Request,
        status: int = 0,
        content: bytes = None,
        headers: dict[str, str] = None,
        encoding: str = 'utf-8') -> Response:
    return Response(request, status=status, content=content, headers=headers, encoding=encoding)



import json as j

from jsonpath_ng import parse

from .selector import Selector
from .. import Response
from ..types import Converter


class Json(Selector):
    def __init__(self,
                 expr: str,
                 selector: Selector = None,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Json, self).__init__(expr,
                                   selector=selector,
                                   multi=multi,
                                   trim=trim,
                                   filter_empty=filter_empty,
                                   format_str=format_str,
                                   converter=converter)
        self._parser = parse(self.expr)

    def do_select(self, content: str, resp: Response) -> list[str]:
        matches = self._parser.find(j.loads(content))
        return [(x.value if isinstance(x.value, str) else j.dumps(x.value)) if x.value is not None else '' for x in
                matches]


def new(expr: str,
        selector: Selector = None,
        *,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Json:
    return Json(expr,
                selector,
                multi=multi,
                trim=trim,
                filter_empty=filter_empty,
                format_str=format_str,
                converter=converter)



from urllib.parse import urlparse, parse_qs

from .selector import Selector
from .. import Response
from ..types import Converter


class Query(Selector):
    def __init__(self,
                 expr: str,
                 *,
                 multi=False,
                 trim=True,
                 filter_empty=True,
                 format_str: str = None,
                 converter: Converter = None):
        super(Query, self).__init__(expr,
                                    multi=multi,
                                    trim=trim,
                                    filter_empty=filter_empty,
                                    format_str=format_str,
                                    converter=converter)

    def do_select(self, content: str, resp: Response) -> list[str]:
        attr = resp.request.queries.get(self.expr, None)
        if attr is None:
            attr = parse_qs(urlparse(resp.request.url).query).get(self.expr, None)
        if attr is None:
            return []
        return [str(c) for c in attr] if isinstance(attr, list) else [str(attr)]


def new(name: str,
        *,
        multi=False,
        trim=True,
        filter_empty=True,
        format_str: str = None,
        converter: Converter = None) -> Query:
    return Query(name,
                 multi=multi,
                 trim=trim,
                 filter_empty=filter_empty,
                 format_str=format_str,
                 converter=converter)



from setuptools import setup, find_packages

setup(
    name="pyoctopus",  # 库的名称
    version="0.2.2",  # 库的版本
    packages=find_packages(),  # 自动查找并包含所有模块
    install_requires=[  # 项目的依赖包
        "Requests",
        "curl_cffi",
        "beautifulsoup4",
        "jsonpath_ng",
        "lxml",
        "tornado",
        "redis",
        "openpyxl",
    ],
    author="yangshoulai",  # 作者信息
    author_email="shoulai.yang@gmail.com",  # 作者邮箱
    description="A simple Python library for web crawler",  # 库的简短描述
    long_description=open("README.md").read(),  # 从 README.md 中读取库的详细描述
    long_description_content_type="text/markdown",
    url="https://github.com/yangsholai/pyoctopus",  # 项目的 GitHub 地址
    classifiers=[  # 分类信息
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",  # 适用的 Python 版本
)



import logging
from typing import Type, List

from ..request import Request
from ..response import Response
from ..selector import select
from ..types import Collector, Processor, R

_logger = logging.getLogger('pyoctopus.processor.extractor')


def new(result_class: Type[R], collector: Collector = None, *args, **kwargs) -> Processor:
    def process(res: Response) -> List[Request]:
        r, links = select(res.text, res, result_class=result_class, *args, **kwargs)
        if r is not None:
            if collector:
                collector(r)
        else:
            _logger.debug(f'No content found from {res}')
        if not links:
            _logger.debug(f'No links found from {res}')
        return links

    return process



import curl_cffi
import requests
from ..request import Request
from ..response import Response
from ..site import Site

_DEFAULT_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
}


def curl_cffi_downloader(request: Request, site: Site) -> Response:
    h = {**_DEFAULT_HEADERS, **site.headers, **request.headers}
    p = {}
    if site.proxy:
        p = {"http": site.proxy, "https": site.proxy}
    r = curl_cffi.request(
        request.method,
        request.url,
        params=request.queries,
        data=request.data,
        headers=h,
        proxies=p,
        timeout=site.timeout,
        impersonate="chrome",
    )
    res = Response(request)
    res.status = r.status_code
    res.content = r.content
    res.headers = {k.lower(): v for k, v in r.headers.items()}
    res.encoding = r.encoding or site.encoding or "utf-8"
    return res


def requests_downloader(request: Request, site: Site) -> Response:
    h = {**_DEFAULT_HEADERS, **site.headers, **request.headers}
    p = {}
    if site.proxy:
        p = {"http": site.proxy, "https": site.proxy}
    r = requests.request(
        request.method,
        request.url,
        params=request.queries,
        data=request.data,
        headers=h,
        proxies=p,
        timeout=site.timeout,
    )
    res = Response(request)
    res.status = r.status_code
    res.content = r.content
    res.headers = {k.lower(): v for k, v in r.headers.items()}
    res.encoding = r.encoding or site.encoding or "utf-8"
    return res
